"""
Tests for the RedisConnection
"""
# Standard
import copy

# Third Party
import pytest

# Local
from oper8.exceptions import ConfigError, PreconditionError
from oper8.test_helpers.helpers import setup_session
from oper8.test_helpers.oper8x_helpers import set_secret_data
from oper8.x.datastores.redis.connection import RedisConnection
from oper8.x.datastores.redis.factory import RedisFactory
from oper8.x.utils import common, constants

## Helpers #####################################################################

sample_base_kwargs = {
    "hostname": "foo.bar.com",
    "port": 1234,
}

sample_auth_kwargs = {
    "auth_secret_name": "test-auth-secret",
    "auth_secret_username_field": "username",
    "auth_secret_password_field": "password",
}

sample_tls_kwargs = {
    "tls_secret_name": "test-tls-secret",
    "tls_secret_cert_field": "ca.pem",
}

sample_uri_kwargs = {
    "uri_secret_name": "test-uri-secret",
    "uri_secret_hostname_field": "hostname",
    "uri_secret_port_field": "port",
}


REDIS_PROVIDED_CONNECTION_DEPLOY_CONFIG_OVERRIDES = {
    constants.SPEC_DATASTORES: {
        RedisFactory.DATASTORE_TYPE: {
            constants.SPEC_DATASTORE_CONNECTION: {
                "uri_secret": sample_uri_kwargs["uri_secret_name"],
                "uri_secret_hostname_field": sample_uri_kwargs[
                    "uri_secret_hostname_field"
                ],
                "uri_secret_port_field": sample_uri_kwargs["uri_secret_port_field"],
                "auth_secret_name": sample_auth_kwargs["auth_secret_name"],
                "auth_secret_username_field": sample_auth_kwargs[
                    "auth_secret_username_field"
                ],
                "auth_secret_password_field": sample_auth_kwargs[
                    "auth_secret_password_field"
                ],
                "tls_secret_name": sample_tls_kwargs["tls_secret_name"],
                "tls_secret_cert_field": sample_tls_kwargs["tls_secret_cert_field"],
            },
            "secondInstance": {
                constants.SPEC_DATASTORE_CONNECTION: {
                    "uri_secret": "second-instance-secret",
                    "uri_secret_hostname_field": "hn",
                    "uri_secret_port_field": "port",
                    "auth_secret_name": "second-instance-secret",
                    "auth_secret_username_field": "un",
                    "auth_secret_password_field": "pw",
                    "tls_secret_name": "second-instance-secret",
                    "tls_secret_cert_field": "crt",
                },
            },
        }
    }
}

sample_auth_secret_kwargs = {
    "auth_username": "itsmeamario",
    "auth_password": "IAmVerySecret",
}

sample_tls_secret_kwargs = {
    "tls_cert": "No cert to be found",
}

sample_uri_secret_kwargs = {
    "hostname": "fake-hostname",
    "port": str(141414),
}

auth_secret_data = {
    sample_auth_kwargs["auth_secret_username_field"]: common.b64_secret(
        sample_auth_secret_kwargs["auth_username"]
    ),
    sample_auth_kwargs["auth_secret_password_field"]: common.b64_secret(
        sample_auth_secret_kwargs["auth_password"]
    ),
}

tls_secret_data = {
    sample_tls_kwargs["tls_secret_cert_field"]: common.b64_secret(
        sample_tls_secret_kwargs["tls_cert"]
    ),
}

uri_secret_data = {
    sample_uri_kwargs["uri_secret_hostname_field"]: common.b64_secret(
        sample_uri_secret_kwargs["hostname"]
    ),
    sample_uri_kwargs["uri_secret_port_field"]: common.b64_secret(
        sample_uri_secret_kwargs["port"]
    ),
}

second_instance_secret_data = {
    "hn": "foo.bar",
    "port": "1234",
    "un": "me",
    "pw": "shhhh",
    "crt": "---BEGIN CERTIFICATE OR SOMETHING---",
}


def set_auth_secret_data(session):
    set_secret_data(
        session=session,
        scoped_name=False,
        name=sample_auth_kwargs["auth_secret_name"],
        data=auth_secret_data,
    )


def set_tls_secret_data(session):
    set_secret_data(
        session=session,
        scoped_name=False,
        name=sample_tls_kwargs["tls_secret_name"],
        data=tls_secret_data,
    )


def set_uri_secret_data(session):
    set_secret_data(
        session=session,
        scoped_name=False,
        name=sample_uri_kwargs["uri_secret_name"],
        data=uri_secret_data,
    )


def set_redis_secrets(session):
    set_auth_secret_data(session)
    set_tls_secret_data(session)
    set_uri_secret_data(session)

    # Quick-n-dirty for the second secret
    set_secret_data(
        session=session,
        scoped_name=False,
        name="second-instance-secret",
        data={k: common.b64_secret(v) for k, v in second_instance_secret_data.items()},
    )


@pytest.fixture
def redis_session():
    override_deploy_configs = REDIS_PROVIDED_CONNECTION_DEPLOY_CONFIG_OVERRIDES
    session = setup_session(deploy_config=override_deploy_configs)
    set_redis_secrets(session)
    yield session


## Tests #######################################################################


def test_get_component_with_provided_connection(redis_session):
    """Test that a provided connection passed in through the CR returns None on
    a get_component call
    """
    component = RedisFactory.get_component(redis_session)
    assert component is None


def test_get_connection_with_provided_connection(redis_session):
    """Test that a provided connection passed in through the CR returns valid
    connection details on get_connection call
    """
    connection = RedisFactory.get_connection(redis_session)
    assert isinstance(connection, RedisConnection)
    assert connection.hostname == sample_uri_secret_kwargs["hostname"]
    assert connection.port == int(sample_uri_secret_kwargs["port"])


def test_get_connection_with_provided_named_second_connection(redis_session):
    """Test that a provided second connection passed in through the CR returns
    valid connection details on get_connection call
    """
    connection = RedisFactory.get_connection(redis_session, "secondInstance")
    assert isinstance(connection, RedisConnection)
    assert connection.hostname == second_instance_secret_data["hn"]
    assert connection.port == int(second_instance_secret_data["port"])


def test_get_connection_string_missing_secrets():
    """Make sure that trying to get the connection string when the secrets are
    not available raises a PreconditionError as Redis needs username / password
    """
    session = setup_session(
        deploy_config=REDIS_PROVIDED_CONNECTION_DEPLOY_CONFIG_OVERRIDES
    )
    conn = RedisConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
    )
    with pytest.raises(PreconditionError):
        conn.get_connection_string()


def test_properties(redis_session):
    """Test that all the properties return correctly"""
    connection = RedisFactory.get_connection(redis_session)
    assert isinstance(connection, RedisConnection)
    assert connection.schema == "rediss"
    # URI
    assert connection.hostname == sample_uri_secret_kwargs["hostname"]
    assert connection.port == int(sample_uri_secret_kwargs["port"])
    # Auth
    assert connection.auth_secret_name == sample_auth_kwargs["auth_secret_name"]
    assert (
        connection.auth_secret_username_field
        == sample_auth_kwargs["auth_secret_username_field"]
    )
    assert (
        connection.auth_secret_password_field
        == sample_auth_kwargs["auth_secret_password_field"]
    )
    # TLS
    assert connection.tls_secret_name == sample_tls_kwargs["tls_secret_name"]
    assert (
        connection.tls_secret_cert_field == sample_tls_kwargs["tls_secret_cert_field"]
    )
    assert connection.tls_enabled


def test_to_dict(redis_session):
    """Test that serializing to a dict works as expected"""
    connection = RedisFactory.get_connection(redis_session)
    conn_dict = connection.to_dict()
    assert conn_dict == {
        "hostname": connection.hostname,
        "port": connection.port,
        "auth_secret_name": connection.auth_secret_name,
        "auth_secret_password_field": connection.auth_secret_password_field,
        "auth_secret_username_field": connection.auth_secret_username_field,
        "tls_secret_name": connection.tls_secret_name,
        "tls_secret_cert_field": connection.tls_secret_cert_field,
    }


def test_get_auth_username_password(redis_session):
    """Make sure get_auth_username_password correctly looks up the contents in
    the cluster
    """
    connection = RedisFactory.get_connection(redis_session)
    un, pw = connection.get_auth_username_password()
    assert un == common.b64_secret_decode(
        auth_secret_data[connection.auth_secret_username_field]
    )
    assert pw == common.b64_secret_decode(
        auth_secret_data[connection.auth_secret_password_field]
    )


def test_get_tls_cert(redis_session):
    """Make sure get_tls_cert correctly looks up the contents in the cluster"""
    connection = RedisFactory.get_connection(redis_session)
    crt = connection.get_tls_cert()
    assert crt == common.b64_secret_decode(
        tls_secret_data[connection.tls_secret_cert_field]
    )


def test_get_connection_string(redis_session):
    """Make sure get_connection_string correctly formats everything into a
    single connection string
    """
    connection = RedisFactory.get_connection(redis_session)
    conn_str = connection.get_connection_string()
    un, pw = connection.get_auth_username_password()
    assert conn_str == "{}://{}:{}@{}:{}".format(
        connection.schema,
        un,
        pw,
        connection.hostname,
        connection.port,
    )


def test_get_tls_disabled():
    """Make sure that the connection handles being set up without TLS"""
    session = setup_session(
        deploy_config=REDIS_PROVIDED_CONNECTION_DEPLOY_CONFIG_OVERRIDES
    )
    set_redis_secrets(session)
    conn = RedisConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
    )
    assert not conn.tls_enabled
    assert not conn.get_tls_cert()


def test_non_int_port(redis_session):
    """Make sure a non-int port raises a ConfigError"""
    bad_port_secret = copy.deepcopy(uri_secret_data)
    bad_port_secret[sample_uri_kwargs["uri_secret_port_field"]] = common.b64_secret(
        "not an int"
    )
    set_secret_data(
        session=redis_session,
        scoped_name=False,
        name=sample_uri_kwargs["uri_secret_name"],
        data=bad_port_secret,
    )
    with pytest.raises(ConfigError):
        RedisFactory.get_connection(redis_session)



"""
Test the rollout manager's functionality
"""

# Standard
from functools import partial
from unittest import mock
import time

# Third Party
import pytest

# First Party
import alog

# Local
from oper8.dag import CompletionState, Node
from oper8.exceptions import ClusterError, PreconditionError, VerificationError
from oper8.rollout_manager import RolloutManager
from oper8.test_helpers.helpers import DummyNodeComponent, library_config, setup_session

################################################################################
## Helpers #####################################################################
################################################################################

log = alog.use_channel("TEST")


def DummyRolloutComponent(name):
    class _DummyRolloutComponent(DummyNodeComponent):
        """Wrapper for DummyNodeComponent that keeps track of which keeps track of which
        phases have completed and in what order
        """

        def __init__(self, *args, deploy_delay=0, verify_delay=0, **kwargs):
            super().__init__(*args, **kwargs)
            self.completed_states = {}
            self.deploy_delay = deploy_delay
            self.verify_delay = verify_delay

            # Wrap the main lifecycle methods
            for name in ["deploy", "verify"]:
                setattr(self, f"_{name}", getattr(self, name))
                setattr(self, name, partial(self._wrapped_member, name))

        def _wrapped_member(self, name, session):

            # We always render here so that we don't have to worry about
            # pre-rendering in the tests
            self.render_chart(session)
            log.debug("Calling [%s] super().%s", self.name, name)
            delay_name = f"{name}_delay"
            if hasattr(self, delay_name):
                time.sleep(getattr(self, delay_name))
            res = getattr(self, f"_{name}")(session)
            if res:
                self.completed_states[name] = time.time()
            return res

        def deploy_completed(self):
            return "deploy" in self.completed_states

        def verify_completed(self):
            return "verify" in self.completed_states

    setattr(_DummyRolloutComponent, "name", name)
    return _DummyRolloutComponent


def get_end_order(components):
    completed_components = [
        comp for comp in components if "verify" in comp.completed_states
    ]
    return sorted(
        completed_components, key=lambda comp: comp.completed_states["verify"]
    )


################################################################################
## RolloutManager Tests ########################################################
################################################################################


class TestRolloutManager:

    #####################
    ## Rollout Success ##
    #####################

    def test_happy_path(self):
        """Test that a rollout manager can be set up with a session that has
        several dependent components and that it runs to completion correctly
        """

        # Set up a session and parent
        session = setup_session()

        # Set up a small set of components which will all pass
        comp_a = DummyRolloutComponent("A")(session)
        comp_b = DummyRolloutComponent("B")(session)
        comp_c = DummyRolloutComponent("C")(session)

        # Add the deps in a simple dependency relation
        session.add_component_dependency(comp_b, comp_a)
        session.add_component_dependency(comp_c, comp_a)

        # Create the rollout manager and run the rollout
        mgr = RolloutManager(session)
        completion_state = mgr.rollout()
        log.debug2(completion_state)

        # Make sure that it succeeded and that all components completed all
        # forward states
        assert completion_state.deploy_completed()
        assert completion_state.verify_completed()
        assert not completion_state.failed()
        for comp in [comp_a, comp_b, comp_c]:
            assert comp.deploy_completed()
            assert comp.verify_completed()

        # Make sure that the components completed in the expected order (A
        # before either B or C)
        end_order = get_end_order([comp_a, comp_b, comp_c])
        assert comp_a in end_order
        assert comp_b in end_order
        assert comp_c in end_order
        assert comp_a == end_order[0]

        # Make sure the completion state looks right
        assert completion_state == CompletionState(
            verified_nodes=[Node("A"), Node("B"), Node("C")],
            unverified_nodes=[],
            failed_nodes=[],
            unstarted_nodes=[],
        )

    def test_rollout_edge_verify_dry_run(self):
        """When performing a rollout in dry_run, test that a custom edge
        verification function is not executed.
        """
        with library_config(dry_run=True):
            session = setup_session()

            # Set two nodes with a custom edge dependency
            #
            #  A <-x- B
            comp_a = DummyRolloutComponent("A")(session)
            comp_b = DummyRolloutComponent("B")(session)

            def broken_verify(*_, **__):
                raise RuntimeError("broken")

            session.add_component_dependency(comp_b, comp_a, broken_verify)

            # Create the rollout manager and run the rollout
            mgr = RolloutManager(session)
            completion_state = mgr.rollout()

            # Make sure the rollout succeeded. This test's goal is to ensure that
            # the rollout terminates correctly, so getting here is sufficient to
            # show success
            assert completion_state.deploy_completed()
            assert completion_state.verify_completed()
            assert not completion_state.failed()

    def test_rollout_edge_verify_non_dry_run(self):
        """When performing a rollout non dry_run, test that a custom edge
        verification function is executed.
        """
        with library_config(dry_run=False):
            session = setup_session()

            # Set two nodes with a custom edge dependency which will return False
            #
            #  A <-x- B
            comp_a = DummyRolloutComponent("A")(session)
            comp_b = DummyRolloutComponent("B")(session)
            session.add_component_dependency(comp_b, comp_a, lambda _: False)

            # Create the rollout manager and run the rollout
            mgr = RolloutManager(session)
            completion_state = mgr.rollout()

            # Make sure the rollout got to the end without exception and that B did
            # not deploy due to the edge dependency
            assert not completion_state.deploy_completed()
            assert not completion_state.verify_completed()
            assert not completion_state.failed()
            assert comp_a in completion_state.verified_nodes
            assert comp_b in completion_state.unstarted_nodes

    def test_happy_path_after_deploy_verify(self):
        """Test that a rollout manager can be set up with a session that has
        several dependent components and that it runs to completion correctly
        when after_deploy and after_verify hooks is configured and passes.
        """

        # Set up a minimal DAG with a single node
        session = setup_session()
        comp = DummyRolloutComponent("A")(session)

        # Create the rollout manager and run the rollout
        after_deploy = mock.Mock(return_value=True)
        after_verify = mock.Mock(return_value=True)
        mgr = RolloutManager(
            session, after_deploy=after_deploy, after_verify=after_verify
        )
        completion_state = mgr.rollout()
        log.debug2(completion_state)

        # Make sure that it succeeded and that all components completed all
        # forward states
        assert completion_state.deploy_completed()
        assert completion_state.verify_completed()
        assert not completion_state.failed()
        assert after_deploy.called
        assert after_verify.called

    #####################
    ## Rollout Failure ##
    #####################

    def test_deploy_throw(self):
        """Test that a throw during deploy is handled properly"""
        session = setup_session()

        # Set up a linear set of components and configure the second-to-last to
        # fail during deploy.
        #
        #    A -> B -x C
        comp_a = DummyRolloutComponent("A")(session)
        comp_b = DummyRolloutComponent("B")(session, deploy_fail=RuntimeError)
        comp_c = DummyRolloutComponent("C")(session)
        comps = [comp_a, comp_b, comp_c]
        for i, comp in enumerate(comps[1:]):
            session.add_component_dependency(comp, comps[i])

        # Create the rollout manager and run the rollout
        mgr = RolloutManager(session)
        mgr.rollout()

        # Make sure the right set of phases were hit
        assert comp_a.deploy_completed()
        assert not comp_a.verify_completed()
        assert not comp_b.deploy_completed()
        assert not comp_b.verify_completed()
        assert not comp_c.deploy_completed()
        assert not comp_c.verify_completed()

    def test_rollout_verify_incomplete(self):
        """Test that a failed verify test is handled properly as not a failure,
        but also not a completion
        """
        session = setup_session()

        # Set up a linear set of components and configure the second-to-last to
        # fail during deploy.
        #
        #    A -> B -x C
        comp_a = DummyRolloutComponent("A")(session)
        comp_b = DummyRolloutComponent("B")(session, verify_fail=True)
        comp_c = DummyRolloutComponent("C")(session)
        comps = [comp_a, comp_b, comp_c]
        for i, comp in enumerate(comps[1:]):
            session.add_component_dependency(comp, comps[i])

        # Create the rollout manager and run the rollout
        mgr = RolloutManager(session)
        completion_state = mgr.rollout()
        assert completion_state.deploy_completed()
        assert not completion_state.verify_completed()

        # Make sure the right set of phases were hit:
        # - All nodes deployed
        # - Only A completed verification
        assert comp_a.deploy_completed()
        assert comp_b.deploy_completed()
        assert comp_c.deploy_completed()
        assert comp_a.verify_completed()
        assert not comp_b.verify_completed()
        assert not comp_c.verify_completed()

        # Make sure the completion state looks right
        assert completion_state == CompletionState(
            verified_nodes=[Node("A")],
            unverified_nodes=[Node("B"), Node("C")],
            failed_nodes=[],
            unstarted_nodes=[],
        )

    def test_rollout_verify_throw(self):
        """Test that a throw during verify is handled properly."""
        session = setup_session()

        # Set up a linear set of components and configure the second-to-last to
        # fail during deploy.
        #
        #    A -> B -x C
        comp_a = DummyRolloutComponent("A")(session)
        comp_b = DummyRolloutComponent("B")(session, verify_fail=RuntimeError)
        comp_c = DummyRolloutComponent("C")(session)
        comps = [comp_a, comp_b, comp_c]
        for i, comp in enumerate(comps[1:]):
            session.add_component_dependency(comp, comps[i])

        # Create the rollout manager and run the rollout
        #
        # NOTE: An unexpected exception in verify is considered a failed node,
        #   so we consider the deploy phase to have failed here as well. This is
        #   maybe a little aggressive because at this point the deploy() for
        #   each component has run successfully, but the alternative is to treat
        #   unexpected failures in verify as identical to programmatic False
        #   return values which feels too loose. Ultimately, this shouldn't
        #   happen!
        mgr = RolloutManager(session)
        completion_state = mgr.rollout()
        assert not completion_state.deploy_completed()
        assert not completion_state.verify_completed()

        # Make sure the right set of phases were hit:
        # - All nodes deployed
        # - Only A completed verification
        assert comp_a.deploy_completed()
        assert comp_b.deploy_completed()
        assert comp_c.deploy_completed()
        assert comp_a.verify_completed()
        assert not comp_b.verify_completed()
        assert not comp_c.verify_completed()

        # Make sure the completion state looks right
        assert completion_state == CompletionState(
            verified_nodes=[Node("A")],
            unverified_nodes=[Node("C")],
            failed_nodes=[Node("B")],
            unstarted_nodes=[],
        )

    def test_after_deploy_false(self):
        """Test that when after_deploy returns False, the rollout does not
        proceed
        """
        session = setup_session()
        comp = DummyRolloutComponent("A")(session)
        after_deploy = mock.Mock(return_value=False)
        mgr = RolloutManager(session, after_deploy=after_deploy)
        completion_state = mgr.rollout()
        log.debug2(completion_state)
        assert completion_state.deploy_completed()
        assert not completion_state.verify_completed()
        assert not completion_state.failed()
        assert after_deploy.called
        assert isinstance(completion_state.exception, VerificationError)

    def test_after_deploy_non_oper8_error(self):
        """Test that when after_deploy raises a non-oper8 error, the rollout is
        considered a failure
        """

        def fail(*_, **__):
            raise RuntimeError("Non-oper8 exception")

        session = setup_session()
        comp = DummyRolloutComponent("A")(session)
        after_deploy = mock.Mock(side_effect=fail)
        mgr = RolloutManager(session, after_deploy=after_deploy)
        completion_state = mgr.rollout()
        log.debug2(completion_state)
        assert completion_state.deploy_completed()
        assert not completion_state.verify_completed()
        assert completion_state.failed()
        assert after_deploy.called
        assert isinstance(completion_state.exception, RuntimeError)

    def test_after_deploy_fatal_error(self):
        """Test that when after_deploy raises a fatal error, the rollout is
        considered a failure
        """

        def fail(*_, **__):
            raise ClusterError("Fatal error")

        session = setup_session()
        comp = DummyRolloutComponent("A")(session)
        after_deploy = mock.Mock(side_effect=fail)
        mgr = RolloutManager(session, after_deploy=after_deploy)
        completion_state = mgr.rollout()
        log.debug2(completion_state)
        assert completion_state.deploy_completed()
        assert not completion_state.verify_completed()
        assert completion_state.failed()
        assert after_deploy.called
        assert isinstance(completion_state.exception, ClusterError)

    def test_after_deploy_non_fatal_error(self):
        """Test that when after_deploy raises a non-fatal error, the rollout is
        considered incomplete
        """

        def fail(*_, **__):
            raise PreconditionError("Expected error")

        session = setup_session()
        comp = DummyRolloutComponent("A")(session)
        after_deploy = mock.Mock(side_effect=fail)
        mgr = RolloutManager(session, after_deploy=after_deploy)
        completion_state = mgr.rollout()
        log.debug2(completion_state)
        assert completion_state.deploy_completed()
        assert not completion_state.verify_completed()
        assert not completion_state.failed()
        assert after_deploy.called
        assert isinstance(completion_state.exception, PreconditionError)

    def test_after_verify_false(self):
        """Test that when after_verify returns false, a VerificationError is
        added to the the completion state
        """
        session = setup_session()
        comp = DummyRolloutComponent("A")(session)
        after_verify = mock.Mock(return_value=False)
        mgr = RolloutManager(session, after_verify=after_verify)
        completion_state = mgr.rollout()
        log.debug2(completion_state)
        assert completion_state.deploy_completed()
        assert not completion_state.verify_completed()
        assert not completion_state.failed()
        assert after_verify.called
        assert isinstance(completion_state.exception, VerificationError)

    def test_after_verify_non_oper8_error(self):
        """Test that when after_verify raises a non-oper8 error, the rollout is
        considered a failure
        """

        def fail(*_, **__):
            raise RuntimeError("Non-oper8 exception")

        session = setup_session()
        comp = DummyRolloutComponent("A")(session)
        after_verify = mock.Mock(side_effect=fail)
        mgr = RolloutManager(session, after_verify=after_verify)
        completion_state = mgr.rollout()
        log.debug2(completion_state)
        assert completion_state.deploy_completed()
        assert not completion_state.verify_completed()
        assert completion_state.failed()
        assert after_verify.called
        assert isinstance(completion_state.exception, RuntimeError)

    def test_after_verify_fatal_error(self):
        """Test that when after_verify raises a fatal error, the rollout is
        considered a failure
        """

        def fail(*_, **__):
            raise ClusterError("Fatal error")

        session = setup_session()
        comp = DummyRolloutComponent("A")(session)
        after_verify = mock.Mock(side_effect=fail)
        mgr = RolloutManager(session, after_verify=after_verify)
        completion_state = mgr.rollout()
        log.debug2(completion_state)
        assert completion_state.deploy_completed()
        assert not completion_state.verify_completed()
        assert completion_state.failed()
        assert after_verify.called
        assert isinstance(completion_state.exception, ClusterError)

    def test_after_verify_non_fatal_error(self):
        """Test that when after_verify raises a non-fatal error, the rollout is
        considered incomplete
        """

        def fail(*_, **__):
            raise PreconditionError("Expected error")

        session = setup_session()
        comp = DummyRolloutComponent("A")(session)
        after_verify = mock.Mock(side_effect=fail)
        mgr = RolloutManager(session, after_verify=after_verify)
        completion_state = mgr.rollout()
        log.debug2(completion_state)
        assert completion_state.deploy_completed()
        assert not completion_state.verify_completed()
        assert not completion_state.failed()
        assert after_verify.called
        assert isinstance(completion_state.exception, PreconditionError)



"""
This module holds the implementation of the DAG constructs used to perform the
dependency management for rollout
"""

# Standard
from functools import partial
from typing import Callable, Optional

# First Party
import alog

# Local
from . import config
from .component import Component
from .dag import CompletionState, DagHaltError, Runner
from .exceptions import Oper8Error, VerificationError
from .session import Session

log = alog.use_channel("ROLMGR")

## Rollout Functions ###########################################################


def deploy_component(session: Session, component: Component) -> bool:
    """Deploy a component given a particular session

    Args:
        session: Session
            The current rollout session
        component: Component
            The component to deploy

    Returns:
        result: bool
            The result of the deploy
    """
    # Do the render
    with alog.ContextTimer(log.debug2, "Render duration for %s", component):
        component.render_chart(session)
        log.debug3(
            "Rendered objects for [%s]: %s",
            component,
            [str(obj) for obj in component.managed_objects],
        )

    # Do the deploy
    with alog.ContextTimer(log.debug2, "Deploy duration for %s: ", component):
        return component.deploy(session)


def disable_component(session: Session, component: Component) -> bool:
    """Disable a component given a particular session

    Args:
        session: Session
            The current rollout session
        component: Component
            The component to disable

    Returns:
        result: bool
            The result of the disable
    """
    # Do the render
    with alog.ContextTimer(log.debug2, "Render duration for %s", component):
        component.render_chart(session)
        log.debug3(
            "Rendered objects for [%s]: %s",
            component,
            [str(obj) for obj in component.managed_objects],
        )

    # Do the deploy
    with alog.ContextTimer(log.debug2, "Disable duration for %s: ", component):
        return component.disable(session)


def verify_component(session: Session, component: Component) -> bool:
    """Verify a component given a particular session

    Args:
        session: Session
            The current rollout session
        component: Component
            The component to verify

    Returns:
        result: bool
            The result of the verify
    """
    # Do the verify
    with alog.ContextTimer(log.debug2, "Verify duration for %s: ", component):
        return component.verify(session)


## RolloutManager ##############################################################


class RolloutManager:
    """This class manages the dependencies needed to roll out a set of nodes"""

    @classmethod
    def run_node(
        cls,
        func: Callable[[Component, Session], bool],
        session: Session,
        component: Component,
        fail_halt_runner=True,
    ):
        """
        Generic function to execute a node during Rollout

        Args:
            func: Callable[[Component,Session], bool]
                The function to be called
            session: Session
                The session that's currently being rolled out
            component: Component
                The component being rolled out
        """
        success = False
        exception = None
        rollout_failed = False

        try:
            success = func(session, component)
            if fail_halt_runner and not success:
                rollout_failed = True

        # If a failure occurred by throwing, treat that the same as an
        # explict failure.
        except Oper8Error as err:
            log.debug("Caught Oper8Error during rollout of [%s]", component)
            success = False
            rollout_failed = err.is_fatal_error
            exception = err
        except Exception as err:  # pylint: disable=broad-except
            log.warning(
                "Caught exception during rollout of [%s]",
                component,
                exc_info=True,
            )
            success = False
            rollout_failed = True
            exception = err

        # If the rollout failed for any reason, raise an exception. This
        # will halt the graph execution.
        if not success:
            log.debug("[deploy] Halting rollout")
            raise DagHaltError(rollout_failed, exception=exception)

        log.debug3("Done with executing node: %s", component)

    def __init__(
        self,
        session: Session,
        after_deploy: Optional[Callable[[Session], bool]] = None,
        after_verify: Optional[Callable[[Session], bool]] = None,
    ):
        """Construct with the fully-populated session for the rollout

        Args:
            session:  Session
                The current session for the reconciliation
            after_deploy:  Optional[Callable[[Session] bool]]
                An optional callback hook that will be invoked after the deploy
                phase completes. The return indicates whether the validation has
                passed.
            after_verify:  Optional[Callable[[Session] None]]
                An optional callback hook that will be invoked after the verify
                phase completes. The return indicates whether the validation has
                passed.
        """
        self._session = session
        self._after_deploy = after_deploy
        self._after_verify = after_verify

        # Read pool size from config
        deploy_threads = config.rollout_manager.deploy_threads
        verify_threads = config.rollout_manager.verify_threads

        # If session is in standalone mode, the recursive deploy -> render in
        # subsystems can cause jsii to fail in some spectacular ways. As such,
        # we force single-threaded execution in standalone mode.
        if config.standalone:
            log.info("Running rollout without threads in standalone mode")
            deploy_threads = 0
            verify_threads = 0

        deploy_node_fn = partial(
            RolloutManager.run_node,
            deploy_component,
            self._session,
        )

        verify_node_fn = partial(
            RolloutManager.run_node,
            verify_component,
            self._session,
            fail_halt_runner=False,
        )

        # Override disabled components with the disable function
        self.disabled_nodes = set()
        for component in session.graph:
            if component.disabled:
                component.set_data(
                    partial(
                        RolloutManager.run_node,
                        disable_component,
                        self._session,
                        component,
                    )
                )

                self.disabled_nodes.add(component)

        self._deploy_graph = Runner(
            "deploy",
            threads=deploy_threads,
            graph=session.graph,
            default_function=deploy_node_fn,
            verify_upstream=not config.dry_run,
        )
        self._verify_graph = Runner(
            "verify",
            threads=verify_threads,
            graph=session.graph,
            default_function=verify_node_fn,
            verify_upstream=not config.dry_run,
        )

    def rollout(  # pylint: disable=too-many-locals,too-many-statements
        self,
    ) -> CompletionState:
        """Perform the rollout

        The logic has four phases:
            1. Deploy Graph: This phase executes the Runner which runs the deploy()
                function for each Component in dependency order. For graph edges
                with configured verification functions, they are also executed
                during this phase.
            2. After Deploy: If configured with an after_deploy hook and (1)
                completed all nodes successfully, this function is called. Any
                raised exceptions indicate that the rollout should not proceed.
            3. Verify Graph: This phase executes the Runner which runs the verify()
                function for each Component in dependency order.
            4. After Verify: If configured with an after_verify hook and (3)
                completed all nodes successfully, this function is called. Any
                raised exceptions indicate that the rollout is not fully
                verified.

        The rollout can terminate in one of three states:
            1. incomplete AND failed: Something unexpected happened and the
                rollout terminated in a fatal state.
            2. incomplete AND NOT failed: One or more nodes did not pass their
                verify steps, but all deploy steps that were attempted
                succeeded.
            3. complete AND NOT failed: The rollout completed all nodes,
                including all verify steps. The managed components are ready to
                take traffic.

        Returns:
            completion_state:  CompletionState
                The final completion state of all nodes in the rollout Runner. This
                is a logical composition of the outputs of the above phases
                based on the termination logic above.
        """

        # NOTE: The Runner is guaranteed to not throw (unless there's a bug!)
        #   so we don't need to wrap run() in a try/except since the except
        #   clause will never catch "expected" errors

        ###########################
        ## Phase 1: Deploy Graph ##
        ###########################
        with alog.ContextTimer(log.trace, "Deploy Graph [%s]: ", self._session.id):
            self._deploy_graph.run()
        deploy_completion_state = self._deploy_graph.completion_state()

        # Log phase 1 results:
        #   * SUCCESS => All Components ran `render()` and `deploy()` without
        #       raising. This is fetched from the `verify_completed()` state
        #       since Components may raise precondition errors resulting in
        #       `deploy_completed()` returning True, indicating that all nodes
        #       ran and none reached a failed state, but not all nodes running
        #       to final completion
        #   * FAILED => Some nodes raised fatal errors
        #   * INCOMPLETE => No errors were raised, but some nodes did not fully
        #       complete without raising
        log.debug3("Deploy completion: %s", deploy_completion_state)
        phase1_complete = deploy_completion_state.verify_completed()
        phase1_failed = deploy_completion_state.failed()
        log.debug(
            "[Phase 1] Deploy result: %s",
            "SUCCESS"
            if phase1_complete
            else ("FAILED" if phase1_failed else "INCOMPLETE"),
        )

        ###########################
        ## Phase 2: After Deploy ##
        ###########################

        phase2_complete = phase1_complete
        phase2_exception = None
        if phase1_complete and self._after_deploy:
            log.debug2("Running after-deploy")
            try:
                phase2_complete = self._after_deploy(self._session)
                if not phase2_complete:
                    phase2_exception = VerificationError(
                        "After-deploy verification failed"
                    )
            except Exception as err:  # pylint: disable=broad-except
                log.debug2("Error caught during after-deploy: %s", err, exc_info=True)
                phase2_complete = False
                phase2_exception = err

        # Log phase 2 results
        log.debug(
            "[Phase 2] After deploy result: %s",
            "SUCCESS"
            if phase2_complete
            else ("FAILED" if phase2_exception else "NOT RUN"),
        )

        ###########################
        ## Phase 3: Verify Graph ##
        ###########################

        # If phase 1 ran without erroring, then run the verify Runner
        phase3_complete = False
        phase3_failed = False
        if not phase1_failed:
            # Configured the verify Runner based off of deployed nodes
            # This way only components that have started will be verified
            deployed_nodes = (
                deploy_completion_state.verified_nodes.union(
                    deploy_completion_state.unverified_nodes
                )
            ) - deploy_completion_state.failed_nodes
            log.debug3("Verify phase running with Nodes: %s", deployed_nodes)

            # Enable/Disable all nodes in verify_graph based on whether they
            # were deployed or not
            for comp in set(self._session.get_components()):
                if comp in deployed_nodes:
                    self._verify_graph.enable_node(comp)
                else:
                    self._verify_graph.disable_node(comp)

                # Disabled components should immediately verify
                if comp in self.disabled_nodes:
                    comp.set_data(lambda *_: True)

            # Run the verify Runner
            with alog.ContextTimer(log.trace, "Verify Graph [%s]: ", self._session.id):
                self._verify_graph.run()
            verify_completion_state = self._verify_graph.completion_state()
            log.debug3("Verify completion: %s", verify_completion_state)
            # Only consider phase3 completed if phase1 and phase2 fully completed
            phase3_complete = (
                verify_completion_state.verify_completed()
                and phase1_complete
                and phase2_complete
            )
            phase3_failed = verify_completion_state.failed()
        else:
            verify_completion_state = CompletionState()

        # Log phase 3 results
        log.debug(
            "[Phase 3] Verify result: %s",
            "SUCCESS"
            if phase3_complete
            else (
                "FAILED"
                if phase3_failed
                else ("INCOMPLETE" if phase2_complete else "NOT RUN")
            ),
        )

        ###########################
        ## Phase 4: After Verify ##
        ###########################

        phase4_complete = phase3_complete
        phase4_exception = None
        if phase3_complete and self._after_verify:
            log.debug("Running after-verify")
            try:
                phase4_complete = self._after_verify(self._session)
                if not phase4_complete:
                    phase4_exception = VerificationError(
                        "Application verification incomplete"
                    )
            except Exception as err:  # pylint: disable=broad-except
                log.debug2("Error caught during after-verify: %s", err, exc_info=True)
                phase4_complete = False
                phase4_exception = err

        # Log phase 4 results
        log.debug(
            "[Phase 4] After deploy result: %s",
            "SUCCESS"
            if phase4_complete
            else ("FAILED" if phase4_exception else "NOT RUN"),
        )

        # Create a final completion state with the "deployed nodes" pulled
        # from the deploy results and the "verified nodes" pulled from the
        # verify results.
        #
        # Verified Nodes: Nodes that made it all the way through the verify
        #   graph
        # Unverified Nodes: Nodes that are "verified" in the deploy graph, but
        #   are unverified in the verify graph or were not run in the verify
        #   graph and did not fail in the verify graph
        # Failed Nodes: Nodes that failed in either graph
        # Unstarted Nodes: Nodes that were unstarted in the deploy graph
        # Exception: Any exception from any of the phases above
        verified_nodes = verify_completion_state.verified_nodes
        failed_nodes = verify_completion_state.failed_nodes.union(
            deploy_completion_state.failed_nodes
        )
        unverified_nodes = (
            (
                deploy_completion_state.verified_nodes.union(
                    deploy_completion_state.unverified_nodes
                ).union(verify_completion_state.unverified_nodes)
            )
            - verified_nodes
            - failed_nodes
        )
        unstarted_nodes = deploy_completion_state.unstarted_nodes
        exception = (
            deploy_completion_state.exception
            or phase2_exception
            or verify_completion_state.exception
            or phase4_exception
        )
        completion_state = CompletionState(
            verified_nodes=verified_nodes,
            unverified_nodes=unverified_nodes,
            failed_nodes=failed_nodes,
            unstarted_nodes=unstarted_nodes,
            exception=exception,
        )
        log.debug2("Final rollout state: %s", completion_state)
        return completion_state



"""
Test the functions in the common utility
"""

# Third Party
import pytest

# First Party
import aconfig
import alog

# Local
from oper8.exceptions import ConfigError
from oper8.test_helpers.helpers import (
    MockDeployManager,
    configure_logging,
    setup_cr,
    setup_session,
)
from oper8.test_helpers.oper8x_helpers import set_object_test_state
from oper8.x.utils import common

configure_logging()
log = alog.use_channel("TEST")


## Unit Tests ##################################################################


def test_get_replicas():

    # Size small
    _app_config = aconfig.Config(
        {"replicas": {"small": {"foo": 1}, "medium": {"foo": 2}}}
    )
    _deploy_config = aconfig.Config({"size": "small"})
    _cr = setup_cr(
        metadata={"namespace": "test", "name": "foo"},
        kind="Deployment",
        apiVersion="apps/v1",
    )
    _dm = MockDeployManager(resources=[_cr])

    session = setup_session(
        app_config=_app_config,
        deploy_config=_deploy_config,
        deploy_manager=_dm,
    )
    # get_replicas when previous state is not present should return valid
    # replica count
    assert common.get_replicas(session, "foo", "foo") == 1

    # Make sure an override is respected when resource is not present
    assert common.get_replicas(session, "foo", "foo", replicas_override=2) == 2

    # Set the state in the cluster and make sure that get_replicas returns
    # the current replication count
    success, changed = set_object_test_state(
        session,
        kind="Deployment",
        name="foo",
        value={
            "apiVersion": "apps/v1",
            "metadata": {
                "labels": common.get_deploy_labels(
                    session,
                )
            },
            "spec": {"replicas": 3},
        },
    )
    assert success
    assert changed
    assert common.get_replicas(session, "foo", "foo") == 3

    # Make sure an override is not used when resource is present
    assert common.get_replicas(session, "foo", "foo", replicas_override=2) == 3

    # Make sure a force=True call returns the value even when the resource
    # exists in the cluster
    assert common.get_replicas(session, "foo", "foo", force=True) == 1

    # Make sure that changing the t-shirt size causes the replicas to be
    # returned
    session.spec.size = "medium"
    assert common.get_replicas(session, "foo", "foo") == 2

    # Size medium: make sure sizes are used as keys correctly
    session = setup_session(
        app_config=_app_config,
        deploy_config=aconfig.Config({"size": "medium"}),
    )
    assert common.get_replicas(session, "foo", "foo") == 2

    # Size large: make sure misconfigured size raises a ConfigError
    session = setup_session(
        app_config=_app_config,
        deploy_config=aconfig.Config({"size": "large"}),
    )
    with pytest.raises(ConfigError):
        common.get_replicas(session, "foo", "foo")


def test_snake_case_to_camelcase():
    test_input = [
        None,
        "string_snake_case",
        {
            "dict_key_snake_case": "dict_value_snake_case",
            "dict_key_snake_case1": None,
        },
    ]
    expected = [
        None,
        "string_snake_case",
        {"dictKeySnakeCase": "dict_value_snake_case", "dictKeySnakeCase1": None},
    ]
    camel_conversion = common.snake_case_to_camelcase(test_input)
    assert camel_conversion == expected
    round_trip = common.camelcase_to_snake_case(camel_conversion)
    assert round_trip == test_input



"""
Test the TLS utility functionality
"""

# Third Party
from cryptography import x509

# First Party
import alog

# Local
from oper8.x.utils import tls

log = alog.use_channel("TEST")


def test_get_subject_valid_type():
    """Make sure the type returned by get_subject is the right type"""
    subject = tls.get_subject()
    assert isinstance(subject, x509.Name)



"""
Tests for the Filter classes
"""
# Local
from oper8.deploy_manager.kube_event import KubeEventType
from oper8.status import ReadyReason, make_application_status
from oper8.test_helpers.pwm_helpers import make_managed_object
from oper8.watch_manager.python_watch_manager.filters.filters import (
    AnnotationFilter,
    CreationDeletionFilter,
    DependentWatchFilter,
    GenerationFilter,
    LabelFilter,
    NoGenerationFilter,
    PauseFilter,
    ResourceVersionFilter,
    SubsystemStatusFilter,
    UserAnnotationFilter,
)

## Helpers #####################################################################


def test_filter_creation_deletion():
    resource = make_managed_object()
    filter = CreationDeletionFilter(resource)

    assert filter.update_and_test(resource, KubeEventType.ADDED)
    assert filter.update_and_test(resource, KubeEventType.DELETED)
    assert filter.update_and_test(resource, KubeEventType.MODIFIED) == None


def test_filter_generation():
    resource = make_managed_object(generation=1)
    filter = GenerationFilter(resource)

    # Assert Generation doesn't care the first time it sees a resource
    assert filter.update_and_test(resource, KubeEventType.MODIFIED) == None
    resource.definition["metadata"]["generation"] = 2
    assert filter.update_and_test(resource, KubeEventType.MODIFIED)
    assert filter.update_and_test(resource, KubeEventType.MODIFIED) == False
    assert filter.update_and_test(resource, KubeEventType.DELETED) == None
    assert filter.update_and_test(resource, KubeEventType.ADDED) == None


def test_filter_no_generation():
    resource = make_managed_object(generation=None)
    filter = NoGenerationFilter(resource)

    # Assert Generation doesn't care the first time it sees a resource
    assert filter.update_and_test(resource, KubeEventType.MODIFIED) == None
    resource.definition["spec"]["changed"] = "value"
    assert filter.update_and_test(resource, KubeEventType.MODIFIED)
    assert filter.update_and_test(resource, KubeEventType.MODIFIED) == False
    assert filter.update_and_test(resource, KubeEventType.DELETED) == None
    assert filter.update_and_test(resource, KubeEventType.ADDED) == None


def test_filter_resource_version():
    resource = make_managed_object()
    filter = ResourceVersionFilter(resource)

    assert filter.update_and_test(resource, KubeEventType.MODIFIED)
    resource = make_managed_object(resource_version="arandomvalue")
    assert filter.update_and_test(resource, KubeEventType.MODIFIED)
    assert filter.update_and_test(resource, KubeEventType.MODIFIED) == False
    assert filter.update_and_test(resource, KubeEventType.ADDED) == False
    assert filter.update_and_test(resource, KubeEventType.DELETED) == None


def test_filter_annotation():
    resource = make_managed_object()
    filter = AnnotationFilter(resource)

    assert filter.update_and_test(resource, KubeEventType.MODIFIED)
    resource = make_managed_object(annotations={"updated": "value"})
    assert filter.update_and_test(resource, KubeEventType.MODIFIED)
    assert filter.update_and_test(resource, KubeEventType.MODIFIED) == False
    resource.definition["spec"]["updated"] = "value"
    assert filter.update_and_test(resource, KubeEventType.MODIFIED) == False

    # Check add and deletes are ignored
    assert filter.update_and_test(resource, KubeEventType.ADDED) == None
    assert filter.update_and_test(resource, KubeEventType.DELETED) == None


def test_filter_user_annotation():
    resource = make_managed_object()
    filter = UserAnnotationFilter(resource)

    assert filter.update_and_test(resource, KubeEventType.MODIFIED)
    resource = make_managed_object(annotations={"updated": "value"})
    assert filter.update_and_test(resource, KubeEventType.MODIFIED)
    assert filter.update_and_test(resource, KubeEventType.MODIFIED) == False
    resource = make_managed_object(
        annotations={"updated": "value", "k8s.io/ingress": "nginx"}
    )
    assert filter.update_and_test(resource, KubeEventType.MODIFIED) == False
    resource.definition["spec"]["updated"] = "value"
    assert filter.update_and_test(resource, KubeEventType.MODIFIED) == False

    # Check add and deletes are ignored
    assert filter.update_and_test(resource, KubeEventType.ADDED) == None
    assert filter.update_and_test(resource, KubeEventType.DELETED) == None


def test_filter_pause():
    resource = make_managed_object()
    filter = PauseFilter(resource)

    assert filter.test(resource, KubeEventType.ADDED)
    resource = make_managed_object(annotations={"oper8.org/pause-execution": "true"})
    assert not filter.test(resource, KubeEventType.ADDED)


def test_subsystem_filter():
    resource = make_managed_object()
    filter = SubsystemStatusFilter(resource)

    assert filter.update_and_test(resource, KubeEventType.ADDED) == None
    resource = make_managed_object(
        status=make_application_status(ready_reason=ReadyReason.IN_PROGRESS)
    )
    assert filter.update_and_test(resource, KubeEventType.MODIFIED)
    resource = make_managed_object(
        status=make_application_status(ready_reason=ReadyReason.STABLE)
    )
    assert filter.update_and_test(resource, KubeEventType.MODIFIED)
    assert not filter.update_and_test(resource, KubeEventType.MODIFIED)


def test_filter_dependent():
    resource = make_managed_object()
    filter = DependentWatchFilter(resource)

    assert not filter.test(resource, KubeEventType.ADDED)
    assert filter.test(resource, KubeEventType.MODIFIED)
    assert filter.test(resource, KubeEventType.DELETED)


def test_filter_label():
    class CustomLabelFilter(LabelFilter):
        labels = {"app": "oper8", "subsystem": "test"}

    resource = make_managed_object()
    filter = CustomLabelFilter(resource)

    assert not filter.test(resource, KubeEventType.ADDED)
    resource = make_managed_object(labels={"app": "oper8", "subsystem": "test"})
    assert filter.test(resource, KubeEventType.ADDED)



"""
Component base class for building larger abstractions off of
"""

# Standard
from typing import Any, Callable, List, Optional, Tuple
import abc
import os

# Third Party
import yaml

# First Party
import aconfig
import alog

# Local
from . import config
from .constants import INTERNAL_NAME_ANOTATION_NAME, TEMPORARY_PATCHES_ANNOTATION_NAME
from .dag import Graph, Node, ResourceNode
from .deploy_manager import DeployMethod
from .exceptions import assert_cluster
from .managed_object import ManagedObject
from .patch import apply_patches
from .session import COMPONENT_VERIFY_FUNCTION, Session
from .utils import abstractclassproperty, sanitize_for_serialization
from .verify_resources import RESOURCE_VERIFY_FUNCTION, verify_resource

log = alog.use_channel("COMP-BASE")


class Component(Node, abc.ABC):
    """
    This file defines the top-level interface for a "Component" in the
    deployment ecosystem. Each Component will ultimately resolve to a Node in
    the deployment execution graph which can be atomically rendered, deployed,
    verified, and if needed reverted.
    """

    @abstractclassproperty
    def name(self):
        """All Components must implement a name class attribute"""

    def __init__(
        self,
        session: Session,
        disabled: bool = False,
    ):
        """Construct with the session for this deployment

        Args:
            session:  Session
                The session that this component will belong to
            disabled:  bool
                Whether or not this component is disabled
        """
        # Ensure that the name property is defined by accessing it and that
        # namespace is inherited from session.
        self.name  # noqa: B018
        self.session_namespace = session.namespace
        self.disabled = disabled

        # Initialize Node with name
        super().__init__(self.name)

        # Register with the session
        # NOTE: This is done before the parent initialization so duplicates can
        #   be caught by the session with a nice error rather than Graph
        log.debug2("[%s] Auto-registering %s", session.id, self)
        session.add_component(self)

        # Initialize the Graph that'll control component rendering
        self.graph = Graph()

        # The list of all managed objects owned by this component
        self._managed_objects = None

    def __str__(self):
        return f"Component({self.name})"

    @property
    def managed_objects(self) -> List[ManagedObject]:
        """The list of managed objects that this Component currently knows
        about. If called before rending, this will be an empty list, so it will
        always be iterable.

        Returns:
            managed_objects:  List[ManagedObject]
                The list of known managed objects
        """
        return self._managed_objects or []

    ## Base Class Interface ####################################################
    #
    # These methods MAY be implemented by children, but contain default
    # implementations that are appropriate for simple cases.
    #
    # NOTE: We liberally use pylint disables here to make the base interface
    #   clear to deriving classes.
    ##

    def build_chart(self, session: Session):  # pylint: disable=unused-argument
        """The build_chart function allows the derived class to add child Charts
        lazily so that they can take advantage of post-initialization
        information.

        Args:
            session:  Session
                The current deploy session
        """

    def verify(self, session):
        """The verify function will run any necessary testing and validation
        that the component needs to ensure that rollout was successfully
        completed.

        Args:
            session:  Session
                The current reconciliation session

        Returns:
            success:  bool
                True on successful deployment verification, False on failure
                conditions
        """
        return self._default_verify(session, is_subsystem=False)

    @alog.logged_function(log.debug2)
    @alog.timed_function(log.debug2)
    def render_chart(self, session):
        """This will be invoked by the parent Application to build and render
        the individual component's chart

        Args:
            session:  Session
                The session for this reconciliation
        """

        # Do the rendering
        self.__render(session)

        # If a working directory is configured, use it
        if config.working_dir:
            rendered_file = self.to_file(session)
            log.debug("Rendered %s to %s", self, rendered_file)

    def update_object_definition(
        self,
        session: Session,  # pylint: disable=unused-argument
        internal_name: str,  # pylint: disable=unused-argument
        resource_definition: dict,
    ):
        """Allow children to inject arbitrary object mutation logic for
        individual managed objects

        The base implementation simply returns the given definition as a
        passthrough

        Args:
            session:  Session
                The session for this reconciliation
            internal_name:  str
                The internal name of the object to update
            resource_definition:  dict
                The dict representation of the resource to modify

        Returns:
            resource_definition:  dict
                The dict representation of the resource with any modifications
                applied
        """
        return resource_definition

    @alog.logged_function(log.debug2)
    @alog.timed_function(log.debug2)
    def deploy(self, session):
        """Deploy the component

        Args:
            session:  Session
                The current reconciliation session

        Returns:
            success:  bool
                True on successful application of the kub state (not
                programmatic verification), False otherwise
        """
        assert (
            self._managed_objects is not None
        ), "Cannot call deploy() before render_chart()"

        # Deploy all managed objects
        for obj in self.managed_objects:
            success, _ = session.deploy_manager.deploy(
                resource_definitions=[obj.definition],
                method=obj.deploy_method or DeployMethod.DEFAULT,
            )
            if not success:
                log.warning("Failed to deploy [%s]", self)
                return False
        return True

    def disable(self, session):
        """Disable the component

        Args:
            session:  Session
                The current reconciliation session

        Returns:
            success:  bool
                True on successful application of the kub state (not
                programmatic verification), False otherwise
        """
        assert (
            self._managed_objects is not None
        ), "Cannot call disable() before render_chart()"

        # Disable all managed objects
        success, _ = session.deploy_manager.disable(
            [obj.definition for obj in self._managed_objects]
        )
        if not success:
            log.warning("Failed to disable [%s]", self)
            return False
        return True

    ## Resource Interface ####################################################
    #
    # These methods offer functionality that children can use to add resources to
    # a components graph
    ##

    def add_resource(
        self,
        name: str,  # pylint: disable=redefined-builtin
        obj: Any,
        verify_function: Optional[RESOURCE_VERIFY_FUNCTION] = None,
        deploy_method: Optional[DeployMethod] = DeployMethod.DEFAULT,
    ) -> Optional[
        ResourceNode
    ]:  # pylint: disable=unused-argument, redefined-builtin, invalid-name
        """The add_resource function allows the derived class to add resources
        to this component to later be rendered

        Args:
            name:  str
                The name of the resource in the Graph
            obj: Any
                An object or dict which can be manipulated into a dict
                representation of the kubernetes resource
        """
        # Sanitize object to enable native support for openapi objects
        obj = sanitize_for_serialization(obj)

        # Add namespace to obj if not present
        obj.setdefault("metadata", {}).setdefault("namespace", self.session_namespace)

        node = ResourceNode(name, obj, verify_function, deploy_method)
        self.graph.add_node(node)
        return node

    def add_dependency(
        self,
        session: Session,
        *components: "Component",
        verify_function: Optional[COMPONENT_VERIFY_FUNCTION] = None,
    ):
        """This add_dependency function sets up a dependency between this component
        and a list of other components. To add a dependency between resources inside
        this component use resource.add_dependency
        Args:
            session:  Session
                The current resource session
            *components:  Components
                Any number of components to be added as a dependency
            verify_function: Optional[verify_function]
                An Optional callable function of the form `def verify(session) -> bool:`
                to use to verify that the dependency has been satisfied. This
                will be used to block deployment of the component beyond
                requiring that the upstream has been deployed successfully.
        """
        for component in components:
            session.add_component_dependency(self, component, verify_function)

    ## Base Class Utilities ####################################################
    #
    # These methods offer shared functionality that children can (and should)
    # use in their implementations
    ##

    @alog.logged_function(log.debug2)
    def to_dict(self, session):
        """
        Render the component and return it as a Dictionary, mainly useful for testing
        :return: Dictionary of the rendered component
        """
        self.__render(session)
        return [obj.definition for obj in self.managed_objects]

    def to_config(self, session):
        """
        Render the component and return it as an AttrDict, mainly useful for testing
        :return: AttrDict of the rendered component
        """

        return [
            aconfig.Config(obj, override_env_vars=False)
            for obj in self.to_dict(session)
        ]

    def to_file(self, session):
        """
        Render the component to disk and return the rendered file path
        :return: str path to rendered file
        """
        assert config.working_dir is not None, "Config must have a working_dir set"

        # If disabled and not dumping disabled components, nothing to do
        if self.disabled and not config.dump_disabled:
            log.debug("Not dumping disabled component: %s", self)
            return None

        # Get the in-memory representation
        objects = self.to_dict(session)

        # Get the output file name and make sure the directory structure exists
        path_parts = [
            config.working_dir,
            ".".join([session.api_version.replace("/", "."), session.kind]).lower(),
            session.name,
        ]
        if self.disabled:
            path_parts.append("DISABLED")
        path_parts.append(self.name)
        output_dir = os.path.join(*path_parts)
        if not os.path.exists(output_dir):
            log.debug2("Creating output dir: %s", output_dir)
            os.makedirs(output_dir)

        # Serialize to a yaml file
        instance_name = session.name
        output_file = os.path.join(output_dir, f"{instance_name}-{self.name}.k8s.yaml")
        log.debug2("Saving %s to %s", self, output_file)
        with open(output_file, "w", encoding="utf-8") as outfile:
            outfile.write("---\n" + yaml.safe_dump_all(objects))

        return output_file

    ## Base Class Implementation Details #######################################
    #
    # These methods provide shared functionality to the base class function
    # implementations and should not be used directly by children
    ##

    @classmethod
    def get_name(cls):  # pylint: disable=arguments-differ
        """Override get_name to support class attribute"""
        return cls.name

    def _default_verify(self, session, is_subsystem=False):
        """The verify function will run any necessary testing and validation
        that the component needs to ensure that rollout was successfully
        completed.

        Args:
            session:  Session
                The current reconciliation session

        Returns:
            success:  bool
                True on successful deployment verification, False on failure
                conditions
        """
        log.debug2("Using default verification for [%s]", self)

        # If this is in dry run mode, we skip verification since this relies on
        # checking for changes in the cluster which won't ever happen
        if config.dry_run:
            log.debug2("No verification to perform in dry_run")
            return True

        # Verify all managed resources
        for resource in self.managed_objects:
            log.debug2("Verifying [%s/%s]", resource.kind, resource.name)
            if not verify_resource(
                kind=resource.kind,
                name=resource.name,
                api_version=resource.api_version,
                session=session,
                is_subsystem=is_subsystem,
                namespace=resource.namespace,
                verify_function=resource.verify_function,
            ):
                log.debug("[%s/%s] not verified", resource.kind, resource.name)
                return False
        log.debug("All managed resources verified for [%s]", self)
        return True

    @staticmethod
    def _preserve_patch_annotation(session, internal_name, resource_definition):
        """This implementation helper checks the current state of the given
        resource and patches the desired state to preserve any temporary patch
        annotations found. This is done so that temporary patches can be applied
        to subsystem CRs managed by a top-level controller.
        """

        # Get the current state of the object
        kind = resource_definition.get("kind")
        api_version = resource_definition.get("apiVersion")
        metadata = resource_definition.get("metadata", {})
        name = metadata.get("name")
        namespace = metadata.get("namespace")
        assert (
            kind is not None and api_version is not None and name is not None
        ), f"Resource {internal_name} missing critical metadata!"
        success, content = session.get_object_current_state(
            kind=kind, name=name, api_version=api_version, namespace=namespace
        )
        assert_cluster(
            success,
            f"Failed to look for current state for [{kind}/{api_version}/{namespace}/{name}]",
        )

        # Look for existing patch annotations
        if content is not None:
            content_meta = content.get("metadata", {})
            patch_anno = content_meta.get("annotations", {}).get(
                TEMPORARY_PATCHES_ANNOTATION_NAME
            )

            # If found, update the resource
            if patch_anno:
                resource_definition.setdefault("metadata", {}).setdefault(
                    "annotations", {}
                )[TEMPORARY_PATCHES_ANNOTATION_NAME] = patch_anno

            # Any time we have metadata changes, we need to include the
            # resourceVersion. It can't hurt to do so, so we will just always do
            # it here if found.
            resource_version = content_meta.get("resourceVersion")
            if resource_version is not None:
                resource_definition["metadata"]["resourceVersion"] = resource_version

            # Make sure any ownerReferences are persisted as well
            owner_refs = content_meta.get("ownerReferences")
            if owner_refs:
                resource_definition["metadata"]["ownerReferences"] = owner_refs

        return resource_definition

    def __build_lazy_charts(self, session):
        """Delegate to the child implementation of build_chart for lazy chart
        construction.
        """
        self.build_chart(session)

    @alog.logged_function(log.debug3)
    def __render(self, session):
        """This is the primary implementation for rendering objects into
        self.managed_objects
        """

        # Short-circuit if already rendered
        if self._managed_objects is not None:
            log.debug2(
                "%s returning %d pre-rendered objects", self, len(self._managed_objects)
            )
            return self.managed_objects

        # Generate name and dict representation of objects
        resource_list = self.__gather_resources(session)

        # Iterate all ApiObject children in dependency order and perform the
        # rendering, including patches and backend modifications.
        self._managed_objects = []
        for name, obj, verify_func, deploy_method in resource_list:
            # Apply any patches to this object
            log.debug2("Applying patches to managed object: %s", name)
            log.debug4("Before Patching: %s", obj)
            obj = apply_patches(name, obj, session.temporary_patches)

            # Make sure any temporary patch annotations that exist already
            # on this resource in the cluster are preserved
            log.debug2("Checking for existing subsystem patches on: %s", name)
            obj = self._preserve_patch_annotation(session, name, obj)

            # Add the internal name annotation if enabled
            if config.internal_name_annotation:
                log.debug2(
                    "Adding internal name annotation [%s: %s]",
                    INTERNAL_NAME_ANOTATION_NAME,
                    name,
                )
                obj.setdefault("metadata", {}).setdefault("annotations", {})[
                    INTERNAL_NAME_ANOTATION_NAME
                ] = name

            # Allow children to inject additional modification logic
            log.debug4("Before Object Updates: %s", obj)
            obj = self.update_object_definition(session, name, obj)

            # Add the resource to the set managed by the is component
            managed_obj = ManagedObject(obj, verify_func, deploy_method)
            log.debug2("Adding managed object: %s", managed_obj)
            log.debug4("Final Definition: %s", obj)
            self._managed_objects.append(managed_obj)

        return self.managed_objects

    def __gather_resources(
        self, session
    ) -> List[Tuple[str, dict, Callable, DeployMethod]]:
        """This is a helper for __render which handles converting resource objects
        into a list of dictionaries.
        """
        # Perform lazy chart creation before finishing rendering
        self.__build_lazy_charts(session)

        # Determine the flattened set of ApiObject children.
        log.debug2("%s populating managed_objects", self)
        topology = self.graph.topology()
        log.debug3("%s topology has %d elements", self, len(topology))
        log.debug4([type(obj) for obj in topology])
        children = [node for node in topology if isinstance(node, ResourceNode)]
        log.debug2("%s found %d ResourceNode children", self, len(children))

        resource_list = []
        for child in children:
            # Construct the managed object with its internal name
            child_name = ".".join([self.name, child.get_name()])
            resource_list.append(
                (child_name, child.manifest, child.verify_function, child.deploy_method)
            )

        return resource_list



"""
Base class interface for a Cloud Object Store (cos) component
"""

# Standard
from abc import abstractmethod
from typing import Optional

# Local
from .... import component
from ..interfaces import Datastore
from .connection import CosConnection

COMPONENT_NAME = "cos"


@component(COMPONENT_NAME)
class ICosComponentBase(Datastore):
    """A COS chart provides access to a single running S3-compatible object
    store instance
    """

    ## Parent Interface ########################################################

    def get_connection(self) -> CosConnection:
        """Get the connection object for this instance"""
        return CosConnection(
            session=self.session,
            hostname=self._get_hostname(),
            port=self._get_port(),
            bucket_name=self._get_bucket_name(),
            auth_secret_name=self._get_auth_secret_name(),
            auth_secret_access_key_field=self._get_auth_secret_access_key_field(),
            auth_secret_secret_key_field=self._get_auth_secret_secret_key_field(),
            tls_secret_name=self._get_tls_secret_name(),
            tls_secret_cert_field=self._get_tls_secret_cert_field(),
            access_key=self._get_access_key(),
            secret_key=self._get_secret_key(),
            tls_cert=self._get_tls_cert(),
        )

    ## Abstract Interface ######################################################
    #
    # This is the interface that needs to be implemented by a child in order to
    # provide the common information that a client will use.
    ##

    ##################
    ## General Info ##
    ##################

    @abstractmethod
    def _get_hostname(self) -> str:
        """Gets the hotsname for the connection. Can be IP address as well.

        Returns:
            hostname:  str
                The hostname (without schema) for the service
        """

    @abstractmethod
    def _get_port(self) -> int:
        """Gets the port where the service is listening

        Returns:
            port:  int
                The port where the service is listening
        """

    @abstractmethod
    def _get_bucket_name(self) -> str:
        """Gets the bucket name for the connection

        Returns:
            bucket_name:  str
                The default bucket name for this instance
        """

    ###############
    ## Auth Info ##
    ###############

    @abstractmethod
    def _get_auth_secret_name(self) -> str:
        """Get the Auth secret name with any scoping applied

        Returns:
            auth_secret_name:  str
                The name of the secret containing the auth secret
        """

    @abstractmethod
    def _get_auth_secret_access_key_field(self) -> str:
        """Get the field form within the auth secret that contains the
        access_key

        Returns:
            access_key_field:  str
                The field within the auth secret that contains the access_key
        """

    @abstractmethod
    def _get_auth_secret_secret_key_field(self) -> str:
        """Get the field form within the auth secret that contains the
        secret_key

        Returns:
            secret_key_field:  str
                The field within the auth secret that contains the secret_key
        """

    @abstractmethod
    def _get_access_key(self) -> Optional[str]:
        """Get the un-encoded content of the access_key if available in-memory.
        Components which proxy an external secret don't need to fetch this
        content from the cluster.

        Returns:
            access_key:  Optional[str]
                The content of the access_key if known
        """

    @abstractmethod
    def _get_secret_key(self) -> Optional[str]:
        """Get the un-encoded content of the secret_key if available in-memory.
        Components which proxy an external secret don't need to fetch this
        content from the cluster.

        Returns:
            secret_key:  Optional[str]
                The content of the secret_key if known
        """

    ##############
    ## TLS Info ##
    ##############

    @abstractmethod
    def _get_tls_secret_name(self) -> Optional[str]:
        """Get the TLS secret name with any scoping applied if tls is enabled

        Returns:
            tls_secret_name:  Optional[str]
                If tls is enabled, returns the name of the secret, otherwise
                None
        """

    @abstractmethod
    def _get_tls_secret_cert_field(self) -> Optional[str]:
        """Get the field from within the tls secret that contains the CA
        certificate a client would need to use to connect

        Returns:
            cert_field:  Optional[str]
                The field within the tls secret where the CA certificate lives
        """

    @abstractmethod
    def _get_tls_cert(self) -> Optional[str]:
        """Get the un-encoded content of the TLS cert if TLS is enabled and
        available in-memory. Components which proxy an external secret don't
        need to fetch this content from the cluster.

        Returns:
            cert_content:  Optional[str]
                The content of the cert if tls is enabled
        """



"""
This module implements custom exceptions
"""

## Base Error ##################################################################


class Oper8Error(Exception):
    """Base class for all oper8 exceptions"""

    def __init__(self, message: str, is_fatal_error: bool):
        """Construct with a flag indicating whether this is a fatal error. This
        will be a static property of all children.
        """
        super().__init__(message)
        self._is_fatal_error = is_fatal_error

    @property
    def is_fatal_error(self):
        """Property indicating whether or not this error should signal a fatal
        state in the rollout
        """
        return self._is_fatal_error


## Fatal Errors ################################################################


class Oper8FatalError(Oper8Error):
    """An Oper8FatalError is one that indicates an unexpected, and likely
    unrecoverable, failure during a reconciliation.
    """

    def __init__(self, message: str = ""):
        super().__init__(message=message, is_fatal_error=True)


class RolloutError(Oper8FatalError):
    """Exception indicating a failure during application rollout"""

    def __init__(self, message: str = "", completion_state=None):
        self.completion_state = completion_state
        super().__init__(message)


class ConfigError(Oper8FatalError):
    """Exception caused during usage of user-provided configuration"""


class ClusterError(Oper8FatalError):
    """Exception caused during chart construction when a cluster operation fails
    in an unexpected way.
    """


## Expected Errors #############################################################


class Oper8ExpectedError(Oper8Error):
    """An Oper8ExpectedError is one that indicates an expected failure condition
    that should cause a reconciliation to terminate, but is expected to resolve
    in a subsequent reconciliation.
    """

    def __init__(self, message: str = ""):
        super().__init__(message=message, is_fatal_error=False)


class PreconditionError(Oper8ExpectedError):
    """Exception caused during chart construction when an expected precondition
    is not met.
    """


class VerificationError(Oper8ExpectedError):
    """Exception caused during resource verification when a desired verification
    state is not reached.
    """


## Assertions ##################################################################


def assert_precondition(condition: bool, message: str = ""):
    """Replacement for assert() which will throw a PreconditionError. This
    should be used when building a chart which requires that a precondition is
    met before continuing.
    """
    if not condition:
        raise PreconditionError(message)


def assert_verified(condition: bool, message: str = ""):
    """Replacement for assert() which will throw a VerificationError. This
    should be used when verifying the state of a resource in the cluster.
    """
    if not condition:
        raise VerificationError(message)


def assert_config(condition: bool, message: str = ""):
    """Replacement for assert() which will throw a ConfigError. This should be
    used when building a chart which requires that certain conditions be true in
    the deploy_config or app_config.
    """
    if not condition:
        raise ConfigError(message)


def assert_cluster(condition: bool, message: str = ""):
    """Replacement for assert() which will throw a ClusterError. This should
    be used when building a chart which requires that an operation in the
    cluster (such as fetching an existing secret) succeeds.
    """
    if not condition:
        raise ClusterError(message)


## Compatibility Exceptions ##################################################################
class Oper8DeprecationWarning(DeprecationWarning):
    """This warning is issued for deprecated APIs"""


class Oper8PendingDeprecationWarning(PendingDeprecationWarning):
    """This warning is issued for APIs that are still supported but will be removed eventually"""



"""__init__ file for leadership election classes. Imports all leadership managers
 and defines a generic helper"""
# Standard
from typing import Type

# Local
from .... import config
from .annotation import AnnotationLeadershipManager
from .base import LeadershipManagerBase
from .dry_run import DryRunLeadershipManager
from .lease import LeaderWithLeaseManager
from .life import LeaderForLifeManager


def get_leader_election_class() -> Type[LeadershipManagerBase]:
    """Get the current configured leadership election"""
    if config.python_watch_manager.lock.type == "leader-for-life":
        return LeaderForLifeManager
    if config.python_watch_manager.lock.type == "leader-with-lease":
        return LeaderWithLeaseManager
    if config.python_watch_manager.lock.type == "annotation":
        return AnnotationLeadershipManager
    if config.python_watch_manager.lock.type == "dryrun":
        return DryRunLeadershipManager
    return DryRunLeadershipManager



"""This file contains common utilities for detecting if a replace operation is required
for a resource
"""
# Standard
from typing import Any, Callable, List

# Third Party
from openshift.dynamic.apply import recursive_list_diff

# First Party
import alog

log = alog.use_channel("DMRPLC_UTILS")


def modified_lists(current_manifest: dict, desired_manifest: dict) -> bool:
    """Helper function to check if there are any differences in the lists of the desired manifest.
    This is required because Kubernetes combines lists which is often not the desired use
    """
    for k in set(current_manifest.keys()).intersection(set(desired_manifest.keys())):
        if isinstance(current_manifest[k], list) and isinstance(
            desired_manifest[k], list
        ):
            if bool(recursive_list_diff(current_manifest[k], desired_manifest[k])):
                return True
        elif isinstance(current_manifest[k], dict) and isinstance(  # noqa: SIM102
            desired_manifest[k], dict
        ):
            if modified_lists(current_manifest[k], desired_manifest[k]):
                return True
    return False


def modified_value_from(current_manifest: Any, desired_manifest: Any) -> bool:
    """Helper function to check if a manifest switched from value to valueFrom. These are mutually
    exclusive thus they require a replace command.
    """
    if isinstance(current_manifest, list) and isinstance(desired_manifest, list):
        iteration_len = min(len(current_manifest), len(desired_manifest))
        for i in range(iteration_len):
            if modified_value_from(current_manifest[i], desired_manifest[i]):
                return True
    if isinstance(current_manifest, dict) and isinstance(desired_manifest, dict):
        if ("value" in current_manifest and "valueFrom" in desired_manifest) or (
            "valueFrom" in current_manifest and "value" in desired_manifest
        ):
            return True
        else:
            for k in set(current_manifest.keys()).intersection(
                set(desired_manifest.keys())
            ):
                if modified_value_from(current_manifest[k], desired_manifest[k]):
                    return True
    return False


REPLACE_FUNCS: List[Callable[[str, str], bool]] = [modified_lists, modified_value_from]


def requires_replace(current_manifest: dict, desired_manifest: dict) -> bool:
    """Function to determine if a resource requires a replace operation instead
    of apply. This can occur due to list merging, or updating envVars

    Args:
        current_manifest (dict): The current manifest in the cluster
        desired_manifest (dict): The desired manifest that should be applied

    Returns:
        bool: If the current manifest requires a replace operation
    """
    for func in REPLACE_FUNCS:
        if func(current_manifest, desired_manifest):
            log.debug4("Manifest requires replace", desired_manifest)
            return True
    return False



"""
This defines the base class for all DeployManager types.
"""

# Standard
from enum import Enum
from typing import Iterator, List, Optional, Tuple
import abc

# Local
from .kube_event import KubeWatchEvent


class DeployMethod(Enum):
    DEFAULT = "default"
    UPDATE = "update"
    REPLACE = "replace"


class DeployManagerBase(abc.ABC):
    """
    Base class for deploy managers which will be responsible for carrying out
    the actual deploy of an Application/Component.
    """

    @abc.abstractmethod
    def deploy(
        self,
        resource_definitions: List[dict],
        manage_owner_references: bool = True,
        method: DeployMethod = DeployMethod.DEFAULT,
    ) -> Tuple[bool, bool]:
        """The deploy function ensures that the resources defined in the list of
        definitions are deployed in the cluster.

        Args:
            resource_definitions:  list(dict)
                List of resource object dicts to apply to the cluster
            manage_owner_references:  bool
                If true, ownerReferences for the parent CR will be applied to
                the deployed object

        Returns:
            success:  bool
                Whether or not the deploy succeeded
            changed:  bool
                Whether or not the deployment resulted in changes
        """

    @abc.abstractmethod
    def disable(self, resource_definitions: List[dict]) -> Tuple[bool, bool]:
        """The disable function ensures that the resources defined in the list of
        definitions are deleted from the cluster

        Args:
            resource_definitions:  list(dict)
                List of resource object dicts to apply to the cluster

        Returns:
            success:  bool
                Whether or not the delete succeeded
            changed:  bool
                Whether or not the delete resulted in changes
        """

    @abc.abstractmethod
    def get_object_current_state(
        self,
        kind: str,
        name: str,
        namespace: Optional[str] = None,
        api_version: Optional[str] = None,
    ) -> Tuple[bool, dict]:
        """The get_current_objects function fetches the current state of a given
        object by name

        Args:
            kind:  str
                The kind of the object to fetch
            name:  str
                The full name of the object to fetch
            namespace:  str
                The namespace to search for the object
            api_version:  str
                The api_version of the resource kind to fetch

        Returns:
            success:  bool
                Whether or not the state fetch operation succeeded
            current_state:  dict or None
                The dict representation of the current object's configuration,
                or None if not present
        """

    @abc.abstractmethod
    def watch_objects(  # pylint: disable=too-many-arguments
        self,
        kind: str,
        api_version: Optional[str] = None,
        namespace: Optional[str] = None,
        name: Optional[str] = None,
        label_selector: Optional[str] = None,
        field_selector: Optional[str] = None,
        resource_version: Optional[str] = None,
    ) -> Iterator[KubeWatchEvent]:
        """The watch_objects function listens for changes in the cluster and returns a
        stream of KubeWatchEvents

        Args:
            kind:  str
                The kind of the object to fetch
            namespace:  str
                The namespace to search for the object
            name:  str
                The name to search for the object
            api_version:  str
                The api_version of the resource kind to fetch
            label_selector:  str
                The label_selector to filter the resources
            field_selector:  str
                The field_selector to filter the resources
            resource_version:  str
                The resource_version the resource must be newer than

        Returns:
            watch_stream: Generator[KubeWatchEvent]
                A stream of KubeWatchEvents generated while watching
        """

    @abc.abstractmethod
    def filter_objects_current_state(  # pylint: disable=too-many-arguments
        self,
        kind: str,
        namespace: Optional[str] = None,
        api_version: Optional[str] = None,
        label_selector: Optional[str] = None,
        field_selector: Optional[str] = None,
    ) -> Tuple[bool, List[dict]]:
        """The filter_objects_current_state function fetches a list of objects
        that match either/both the label or field selector
        Args:
            kind:  str
                The kind of the object to fetch
            namespace:  str
                The namespace to search for the object
            api_version:  str
                The api_version of the resource kind to fetch
            label_selector:  str
                The label_selector to filter the resources
            field_selector:  str
                The field_selector to filter the resources

        Returns:
            success:  bool
                Whether or not the state fetch operation succeeded
            current_state:  List[dict]
                A list of  dict representations for the objects configuration,
                or an empty list if no objects match
        """

    @abc.abstractmethod
    def set_status(  # pylint: disable=too-many-arguments
        self,
        kind: str,
        name: str,
        namespace: Optional[str],
        status: dict,
        api_version: Optional[str] = None,
    ) -> Tuple[bool, bool]:
        """Set the status for an object managed by oper8

        Args:
            kind:  str
                The kind of the object ot fetch
            name:  str
                The full name of the object to fetch
            namespace:  Optional[str]
                The namespace to search for the object. If None search cluster wide
            status:  dict
                The status object to set onto the given object
            api_version:  str
                The api_version of the resource to update

        Returns:
            success:  bool
                Whether or not the state fetch operation succeeded
            changed:  bool
                Whether or not the status update resulted in a change
        """



"""
Custom logging formats that contain more detailed oper8 logs
"""

# First Party
from alog import AlogJsonFormatter
import alog

log = alog.use_channel("CTRLR")


class Oper8JsonFormatter(AlogJsonFormatter):
    """Custom Log Format that extends AlogJsonFormatter to add multiple
    oper8 specific fields to the json. This includes things like identifiers
    of the resource being reconciled, reconciliationId, and thread information
    """

    _FIELDS_TO_PRINT = AlogJsonFormatter._FIELDS_TO_PRINT + [
        "process",
        "thread",
        "threadName",
        "kind",
        "apiVersion",
        "resourceVersion",
        "resourceName",
        "reconciliationId",
    ]

    def __init__(self, manifest=None, reconciliation_id=None):
        super().__init__()
        self.manifest = manifest
        self.reconciliation_id = reconciliation_id

    def format(self, record):
        if self.reconciliation_id:
            record.reconciliationId = self.reconciliation_id

        if resource := getattr(record, "resource", self.manifest):
            record.kind = resource.get("kind")
            record.apiVersion = resource.get("apiVersion")

            metadata = resource.get("metadata", {})
            record.resourceVersion = metadata.get("resourceVersion")
            record.resourceName = metadata.get("name")

        return super().format(record)
